/* Agent Tcl
   Bob Gray
   9 December 1997

   This file implements the Agent_ServerMain procedure.

   Copyright (c) 1995-1997, Robert S. Gray Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#include "platPorting.h"  
#include "tcl.h"
#include "agentTcl.h"
#include "genServerInterface.h"
#include "restrictTcl.h"
#include "tcpipTcl.h"
#include "tclUtility.h"

/*
 * Maximum number of tries and maxium sleep period when sending result
 */

const UINT_32 MAXIMUM_TRIES = 10;
const UINT_32 MAXIMUM_SLEEP = 300;

/* Agent_ServerMain

   Purpose: Execute agents that arrive through the server

     Input: appInitProc = Tcl initialization procedure
 
    Output: The procedure should never return.
*/

void Agent_ServerMain (Tcl_AppInitProc *appInitProc)
{
    int code;
    char *result;
    Tcl_Interp *interp;

#ifdef FIX
	/* version code currently appears in two places */
	/* (here and in Agent_AppInit in tclAppInit.cc) */
#endif

#ifndef LANGUAGE_NAME
#define LANGUAGE_NAME "STATE-TCL"
#endif

    Version version (LANGUAGE_NAME, TCL_MAJOR_VERSION, TCL_MINOR_VERSION, Version::e_PRODUCTION);
    AGENT_TCL *agent = new AGENT_TCL (version);

	/* create the new agent */ 

    if (agent -> agentRestore () != AGENT_OK) {
	exit (ServerInterface::e_CAN_NOT_RESTORE_STATE);
    }

	/* initialize the interpreter */

    interp = agent -> get_interp ();
    result = my_strcpy (interp -> result);

    if ((code = (*appInitProc)(interp)) != TCL_OK) {
	Tcl_AddErrorInfo (interp, "\n    (*** initialization script ***)");
        Agent_SendResult (interp, MAXIMUM_TRIES, MAXIMUM_SLEEP, code);
 	Tcl_DeleteInterp (interp);
	Tcl_Exit (0);
    }

    Tcl_SetResult (interp, result, TCL_DYNAMIC);

	/* execute the agent */

    code = agent -> agentRun ();

	/* send the result to the root agent and exit */

    Agent_SendResult (interp, MAXIMUM_TRIES, MAXIMUM_SLEEP, interp -> resultCode);
    Tcl_DeleteInterp (interp);
    Tcl_Exit (0);
    abort_with_message ("should never get here");   // should never get here
}
