/* Agent Tcl
   Bob Gray
   10 January 1996

   genAgent.h

   This file defines the agent routines and data structures.

   Revision information:

   $Id: genAgent.h,v 1.19 1998/02/19 22:08:29 rgray Exp $

   $Log: genAgent.h,v $
   Revision 1.19  1998/02/19 22:08:29  rgray
   compiles on SGI with CC
   Thu Feb 19 17:08:29 EST 1998

   Revision 1.18  1998/01/13 02:54:37  rgray
   moved agent_accept into the core
   Mon Jan 12 21:54:37 EST 1998

   Revision 1.17  1998/01/13 00:26:22  rgray
   moved agent_meet into the core
   Mon Jan 12 19:26:22 EST 1998

   Revision 1.16  1998/01/10 19:37:21  rgray
   improved Java masks and nearly finished Python meetings
   Sat Jan 10 14:37:21 EST 1998

   Revision 1.15  1998/01/04 20:35:00  rgray
   finished mask handling in Agent Python and conolidated all of the core's "select" loops into a single class
   Sun Jan  4 20:34:59 GMT 1998

   Revision 1.14  1997/12/27 18:48:39  rgray
   good start on Agent Python
   Sat Dec 27 18:48:39 GMT 1997

   Revision 1.13  1997/12/21 17:22:30  rgray
   added fork command to Agent Java (and fixed about half a dozen bugs in the Sumatra state capture code -- there is still at least one bug remaining (see newpack.c))
   Sun Dec 21 17:22:30 GMT 1997

   Revision 1.12  1997/12/20 08:38:21  rgray
   checkin
   Sat Dec 20 08:38:21 GMT 1997

   Revision 1.11  1997/12/18 21:11:45  rgray
   removed the AGENT_NOT_ROOT error code
   Thu Dec 18 21:11:45 GMT 1997

   Revision 1.10  1997/12/17 18:42:34  rgray
   added AgentFactory (but it's not yet complete)
   Wed Dec 17 18:42:34 GMT 1997

   Revision 1.9  1997/12/16 23:34:34  rgray
   agent temp directory now passed down to agent; added excludeClasses argument to unpack_the_state
   Tue Dec 16 23:34:34 GMT 1997

   Revision 1.8  1997/12/15 17:52:44  rgray
   runtime exceptions in a migrating Java agent are now caught and sent back to the root agent
   Mon Dec 15 17:52:44 GMT 1997

   Revision 1.7  1997/12/11 14:39:42  rgray
   restoreState now takes arguments that make sense
   Thu Dec 11 14:39:42 GMT 1997

   Revision 1.6  1997/12/09 21:49:20  rgray
   eliminated the need for RESTRICT_TRAP
   Tue Dec  9 21:49:20 GMT 1997

   Revision 1.5  1997/12/09 10:02:26  rgray
   code cleanup
   Tue Dec  9 10:02:26 GMT 1997

   Revision 1.4  1997/12/06 13:51:30  rgray
   server now resolves IP addresses after it forks (so that we do not delay the machine startup process)
   Sat Dec  6 13:51:30 GMT 1997

   Revision 1.3  1997/11/08 17:30:07  rgray
   got things ready for the new Scheme
   Sat Nov  8 12:30:07 EST 1997

   Revision 1.2  1997/10/30 14:20:34  rgray
   renamed class MASKS as class AgentMasks (in a continuing effort to get back to a consistent naming scheme)
   Thu Oct 30 09:20:33 EST 1997

   Copyright (c) 1995-1998, Bob Gray, Dartmouth College

   See the file "agent.terms" for information on usage and redistribution
   of this file and for a DISCLAIMER OF ALL WARRANTIES.
*/

#ifndef _GEN_AGENT_H
#define _GEN_AGENT_H

#ifndef NO_PRAGMAS
#pragma interface
#endif

#ifndef _PLAT_EXCLUSION_H
#include "platExclusion.h"	  // MonitorLock
#endif
#ifndef _PLAT_SIGIO_H
#include "platSigio.h"		  // SigioTrap
#endif
#ifndef _PLAT_SIZES_H
#include "platSizes.h"		  // UINT_32
#endif
#ifndef _PLAT_SIGTERM_H
#include "platSigterm.h"	  // SigtermTrap
#endif
#ifndef _GEN_DSTRING_H
#include "genDString.h"		  // DynamicString
#endif
#ifndef _GEN_SERVER_FLAGS_H
#include "genServerInterface.h"	  // SubmissionTypes and ServerErrorcodes
#endif
#ifndef _GEN_VERSION_H
#include "genVersion.h"		  // Version
#endif
#ifndef _TRUEFALSE_H
#include "truefalse.h"		  // BOOLEAN
#endif

    /* forward declarations */

class AgentId;
class AgentIO;
class AgentConnectionProxy;
class AgentFactory;
class AGENT_INFO;
class DynamicBuffer;
class FileCallbacks;
class ID_LIST;
class INCOMING_MESSAGE;
class INCOMING_MEETING;
class INCOMING_EVENT;
class LOCATION;
class MachineId;
class NAME_LIST;
class AgentMasks;
class AgentMeetings;
class RESTRICT;
class SECURITY;
struct timeval;

    /* return codes from the agent routines */

const int AGENT_BEG_GENERAL		= 0;
const int AGENT_OK			= 0;
const int AGENT_ALREADY_REG		= 1;
const int AGENT_NO_SERVER		= 2;
const int AGENT_BAD_RESPONSE	 	= 3;
const int AGENT_SERVER_ERROR	 	= 4;
const int AGENT_NOT_REG		 	= 5;
const int AGENT_PROGRAM_ERROR		= 6;
const int AGENT_TIMEOUT		 	= 7;
const int AGENT_LOST_BACKGROUND	 	= 8;
const int AGENT_NONEXISTENT		= 9;
const int AGENT_SELF		 	= 10;
const int AGENT_UNKNOWN		 	= 11;
const int AGENT_BAD_STATE		= 12;
const int AGENT_SAME		 	= 13;
const int AGENT_PENDING_EVENT	 	= 14;
const int AGENT_NO_EVENT	 	= 15;
const int AGENT_FAILED_MESSAGE	 	= 16;
const int AGENT_ENCRYPT		 	= 17;
const int AGENT_DECRYPT		 	= 18;
const int AGENT_SERVER_SOURCE	 	= 19;
const int AGENT_SERVER_INTERP	 	= 20;
const int AGENT_SERVER_SYSTEM	 	= 21;
const int AGENT_SERVER_INVALID	 	= 22;
const int AGENT_TOO_MANY_MEETINGS	= 23;
const int AGENT_SERVER_LOST		= 24;
const int AGENT_NAME_USED	 	= 25;
const int AGENT_NAME_RESERVED		= 26;
const int AGENT_BAD_MEETING		= 27;
const int AGENT_CANT_ACCEPT		= 28;
const int AGENT_UNKNOWN_MEETING		= 29;
const int AGENT_CANT_REJECT		= 30;
const int AGENT_CANT_SEND		= 31;
const int AGENT_MEET_BROKEN		= 32;
const int AGENT_CANT_RECEIVE 		= 33;
const int AGENT_NEXT_FILE		= 34;
const int AGENT_CANT_READ		= 35;
const int AGENT_NEXT_MESSAGE		= 36;
const int AGENT_CANT_WRITE		= 37;
const int AGENT_MEET_SELF		= 38;
const int AGENT_NO_JUMPS		= 39;
const int AGENT_SERVER_MAX_AGENTS	= 40;
const int AGENT_SERVER_MAX_ANON         = 41;
const int AGENT_SERVER_LIAR		= 42;
const int AGENT_SERVER_RBEGIN		= 43;
const int AGENT_SERVER_ANON_RBEGIN	= 44;
const int AGENT_SERVER_RCOMM		= 45;
const int AGENT_SERVER_RINFO		= 46;
const int AGENT_SERVER_ANON_RINFO	= 47;
const int AGENT_NO_CHILDREN		= 48;
const int AGENT_SERVER_LOAD		= 49;
const int AGENT_SERVER_TOO_MANY_EVENTS	= 50;
const int AGENT_CANT_FORCE		= 51;
const int AGENT_MEET_REJECTED		= 52;
const int AGENT_END_GENERAL		= 52;

    /* callback class for incoming communication */

class IncomingCallback: public SigioTrap
{
private:

	/* callback */

    void sigioAlarm (int /*fd*/, int /*flags*/) {
	incomingAlarm ();   
    }

protected:

	/* constructor and destructor */

    IncomingCallback (void);
    virtual ~IncomingCallback ();

	/* callback when there might be an incoming item */

    virtual void incomingAlarm (void) = 0;

public:

	/* handle incoming items */

    void handleIncomingItems (void) {
	SigioHandler::checkAndCallTraps ();
    }
};

    /* an agent */

class AGENT: public SigtermTrap
{
public:

	/* 
	 * types of AGENT_END calls
	 */

    enum EndType {
	e_END_NORMAL	= 0,
	e_END_CLEANUP   = 1
    };

	/*
	 * types of AGENT_JUMP calls
	 */

    enum JumpType {
	e_JUMPING	= 0,
	e_FORKING	= 1
    };

	/* 
	 * source of the agent 
	 */

    enum AgentSource {
	e_LOCAL_AGENT	 = 0,	// agent started on the local machine
	e_INCOMING_AGENT = 1	// agent arrived from another machine 
    };

	/* 
	 * location of the agent 
	 *
	 * The meaning of the two constants is slightly different than their
	 * names indicate.  e_AT_HOME means that the agent has *never* left
	 * its home machine, i.e., the home and local identifications of the
	 * agent are the same.  e_AWAY_FROM_HOME means that the agent has
	 * left home at some point in its past.  Thus, e_AWAY_FROM_HOME
	 * does not actually indicate whether the agent is currently on or
	 * away from its home machine.
	 *
#ifdef FIX_LATER
         * Once make_root is eliminated (in favor of an optional argument to
         * submit and fork that specifies which agent should receive the child
         * agent's result), AgentSource and AgentLocation will provide the same
         * information and AgentLocation will be removed.
#endif
         */

    enum AgentLocation {
	e_AT_HOME	 = 0,	// agent has never left home 
	e_AWAY_FROM_HOME = 1	// agent has left home
    };

	/*
	 * position of the agent within the parent-child hierarchy
	 */

    enum AgentRelation {
	e_CHILD_AGENT	= 0,	// agent is a child agent
	e_ROOT_AGENT	= 1	// agent is a root agent
    };

	/*
	 * whether agent is registed with a server
	 */

    enum AgentRegister {
	e_REGISTERED	 = 0,	// agent is registered with a server
	e_NOT_REGISTERED = 1	// agent is not registered with a server 
    };

private:

	/* 
	 * lock to provide mutual exclusion 
	 */

    MonitorLock monitor;

private:

	/* convert a MESSAGE code into an AGENT code */

    int messageCodeToAgentCode 
	(int code, AgentConnectionProxy *connection);

	/* convert a SERVER code into an AGENT code */

    int serverCodeToAgentCode
	(ServerInterface::ServerErrorcodes code);

	/* send an information request */

    int agentGenericInfo 
	(UINT_8 flag, struct timeval delay, AgentId *id, void *&list);

	/* send a meeting acceptance */

    int sendAcceptanceMessage 
	(struct timeval stop, const AgentId &id, UINT_32 handle, 
	 UINT_32 remoteHandle, UINT_32 port);

	/* handle a SIGTERM */

    SigtermHandler *sigtermHandler;
    virtual void sigtermAlarm (void);	// inherited from SigtermTrap

protected:

        /* constructor and virtual destructor */

    AGENT (const Version &version, const AgentFactory &factory, AGENT::AgentSource agentSource);
    virtual ~AGENT();

	/* I/O */      

    AgentIO *io;

	/* network location of the agent */

    LOCATION *location; 

	/* "filters" for incoming messages, events and meeting requests */

    AgentMasks *masks;	

	/* pending and active meetings */

    AgentMeetings *meetings;

	/* permits */

    RESTRICT *restrict;

	/* security characteristics */

    SECURITY *security;

	/* core and interpreter versions */

    Version coreVersion;
    Version interpreterVersion;

	/* capture the current state of the agent */

    virtual int captureState 
	(JumpType jumpType, DynamicBuffer& buffer) = 0;

	/* restore the state of an agent */

    virtual int restoreState
	(ServerInterface::SubmissionTypes type, const DynamicBuffer& buffer) = 0;

public:

	/* get the MonitorLock */

    MonitorLock &getMonitorLock (void) {
	return (monitor);
    }

	/* get the I/O structure */

    AgentIO *get_io (void) const {
	return (io);
    }

	/* get network location structure */

    LOCATION *get_location (void) const { 
	return (location); 
    } 

	/* get masks structure */

    AgentMasks *get_masks (void) const { 
	return (masks); 
    } 

	/* get AgentMeetings structure */

    AgentMeetings *get_meetings (void) const {
	return (meetings);
    } 

	/* get the permit structure */

    RESTRICT *get_restrict (void) const {
	return (restrict);
    } 

	/* get the security structure */

    SECURITY *get_security (void) const {
	return (security);
    } 

	/* get the core version */

    Version getCoreVersion (void) const {
	return (coreVersion);
    }

	/* get and set the interpreter version */ 

#ifdef FIX_LATER
	/* should these be protected? */
#endif

    void setInterpVersion (Version &version) {
	interpreterVersion = version;
    }

    Version getInterpVersion (void) const {
	return (interpreterVersion);
    }

	/* get the IP address of the host */

    int agentIp (struct timeval delay, MachineId &machine); 

	/* restore an incoming agent */

    int agentRestore (void);

	/* register an agent with the server */

    int agentBegin (struct timeval delay, const MachineId& server);

	/* remove an agent from the server tables */

    int agentEnd (struct timeval delay, enum EndType endType);

	/* change the name of the agent */

    int agentName (struct timeval delay, char *name);

	/* turn the agent into a root agent */

    int agentRoot (void);

        /* send a message to another agent or receive a message */

    int agentSendMessage 
	(struct timeval delay, const AgentId &id, UINT_32 code, char *string);

    int agentReceiveMessage 
	(struct timeval delay, INCOMING_MESSAGE *&message);

	/* request, receive and accept meetings */

    int agentRequestMeeting 
	(struct timeval delay, const AgentId &id, UINT_32 &handle);

    int agentReceiveMeeting
	(struct timeval delay, INCOMING_MEETING *& meeting);

    int agentAcceptMeeting
	(struct timeval delay, UINT_32 handle, AgentId &id);

    int agentRejectMeeting
	(struct timeval delay, UINT_32 handle, AgentId &id);

    int agentMeetingReceive
	(struct timeval delay, UINT_32 handle, UINT_32 &code, char *&string);

    int agentMeetingGetfile
	(struct timeval delay, UINT_32 handle, FileCallbacks &callbacks,
	 UINT_32 &bytes);

    int agentMeetingSend
	(UINT_32 handle, UINT_32 code, char *string);

    int agentMeetingSendfile
	(UINT_32 handle, FileCallbacks &callbacks, UINT_32 bytes);

    int agentCloseMeeting
	(UINT_32 handle);

    int agentMeet
	(struct timeval delay, const AgentId &id, UINT_32 &handle);

    int agentAccept
	(struct timeval delay, UINT_32 handle, AgentId &id);

	/* send an event to another agent or receive an event */ 

    int agentSendEvent 
	(struct timeval delay, const AgentId &id, char *tag, char *string);

    int agentReceiveEvent 
	(struct timeval delay, INCOMING_EVENT *&event);

	/* select on events, messages, meeting requests and connected */
	/* meetings                                                   */

    int agentSelect
	(struct timeval delay, UINT_32 numHandles, UINT_32 *handlesIn,
	 UINT_8 *handlesOut, UINT_8 *itemsIn, UINT_8 *itemsOut);

	/* submit, fork and jump */

    int agentSubmit
	(struct timeval delay, const MachineId &machine, Version *version, 
	 const DynamicBuffer &script, AgentId *&id);

    int agentJump
	(struct timeval delay, const MachineId &machine, AgentId *&id, 
	 JumpType jumpType);

    int agentDisk
	(FileCallbacks &callbacks);

    int agentTransfer
	(struct timeval delay, const MachineId &machine, FileCallbacks &callbacks,
	 AgentId *&id); 

	/* forcibly terminate an agent */

    int agentForce
	(struct timeval delay, AgentId *id, AgentId *&forcedId);

	/* sleep */      

    int agentSleep 
	(struct timeval delay);

	/* agent information */

    int agentGetInfo
	(struct timeval delay, AgentId *id, AGENT_INFO *&info);

    int agentGetNames
	(struct timeval delay, const MachineId &id, NAME_LIST *&list);

    int agentGetIds 
	(struct timeval delay, const MachineId &id, ID_LIST *&list);

    int agentStatus
	(struct timeval delay, int type, int provideEvent, AgentId *id);

	/* random numbers -- cryptographically secure with help from server */

    void agentSeed
	(UINT_32 seed);

    UINT_32 agentRandom
	(void);

	/* convert an agent error code into a *short* or *long* string */

    DynamicString agentCodeToShortString 
	(int code);

    DynamicString agentCodeToLongString
	(int code, const DynamicString& name);
};

#endif    /* _GEN_AGENT_H */
